import pytest

from xml_serializer.converter_types import String, Integer, Float, Boolean


valid_test_data = [
    (
        String,
        "test string",
        "test string"
    ),
    (
        Integer,
        "123",
        123
    ),
    (
        Float,
        "123.123",
        123.123
    ),
    (
        Boolean,
        "true",
        True
    ),
    (
        Boolean,
        "false",
        False
    ),
]


@pytest.mark.parametrize('converter_type, payload, expected_response', valid_test_data)
def test_valid_cases(converter_type, payload, expected_response):
    converter = converter_type()

    assert converter.convert(payload) == expected_response



from typing import Union
from xml.etree.ElementTree import Element as xml_tag

from .abstract_type import AbstractType


class BaseField:
    def __init__(self, field_name: str, name: str = None):
        self.field_name = field_name

        if not name:
            name = field_name

        self.name = name


class Tag(BaseField):
    """
    Describes a tag in the serialization scheme
    """
    pass


class TagAttr(BaseField):
    """
    Describes a tag attribute in the serialization scheme
    """
    pass


def xml_serialize(schema: dict, tag: xml_tag):
    """
    Serialize xml by schema

    :param schema: schema for serializing
    :param tag: etree element
    :return: serialized xml (to python dict)
    """
    response = {}

    schema_tag, inner_schema = get_schema_tag_and_inner_schema(schema)

    serialized_xml = serialize_by_inner_schema(inner_schema, tag)

    response[schema_tag.name] = serialized_xml

    return response


def serialize_by_inner_schema(schema: dict, tag: xml_tag):
    """
    Use schema without current tag.

    :param schema: schema for serializing
    :param tag: etree element
    :return: serialized xml (to python dict) without current tag
    """
    xml_element = XMLElement(schema, tag)
    response = xml_element.serialize()

    return response


def get_schema_tag_and_inner_schema(schema: dict) -> (Tag, Union[list, dict]):
    schema_items = list(schema.items())

    if len(schema_items) > 1:
        raise ValueError("You can't parse multiple tags without the root tag!")

    schema_tag = schema_items[0][0]
    inner_schema = schema_items[0][1]

    return schema_tag, inner_schema


class XMLElement:
    def __init__(self, schema: dict, tag: xml_tag) -> dict:
        self.schema = schema
        self.tag = tag

    def serialize(self):
        response = {}

        tag_descriptions, tag_attr_descriptions = self.get_field_descriptions()

        if tag_attr_descriptions:
            tag_attrs = self.fill_tag_attrs(
                tag_attr_descriptions
            )
            if tag_attrs:
                response.update(tag_attrs)

        if tag_descriptions:
            tags = self.fill_tags(
                tag_descriptions
            )
            if tags:
                response.update(tags)

        return response or None

    def get_field_descriptions(self) -> (dict, dict):
        tag_descriptions = {}
        tag_attr_descriptions = {}

        for key, value in self.schema.items():
            is_tag = isinstance(key, Tag)

            if is_tag:
                tag_descriptions[key] = value

            else:
                tag_attr_descriptions[key] = value

        return tag_descriptions, tag_attr_descriptions

    def fill_tag_attrs(self, fields_description):
        response = {}

        for key, converter_type in fields_description.items():
            value = self.tag.get(key.field_name)

            if converter_type and value:
                value = converter_type.convert(value)

            response[key.name] = value

        return response

    def fill_tags(self, fields_description: dict):
        children_tags_with_nesting = self.get_children_tags_with_nesting(
            fields_description
        )

        response = {
            key.name: None for key in fields_description
        }

        for children_tag, schema_tag in children_tags_with_nesting.items():
            current_schema_item = self.schema[schema_tag]

            item_type = fields_description[schema_tag]

            if isinstance(item_type, dict):
                response[schema_tag.name] = serialize_by_inner_schema(
                    current_schema_item,
                    children_tag
                )

            elif isinstance(item_type, list):
                self.list_serialize(
                    current_schema_item,
                    children_tag,
                    schema_tag,
                    response
                )

            elif isinstance(item_type, AbstractType):
                response[schema_tag.name] = item_type.convert_method(children_tag)

            else:
                raise TypeError("The wrong type has been set!")

        return response

    def get_children_tags_with_nesting(self, fields_description: dict):
        tag_names = {tag.field_name: tag for tag in fields_description}

        return {
            tag: tag_names[tag.tag] for tag in self.tag
            if tag.tag in tag_names
        }

    @staticmethod
    def list_serialize(
        current_schema_item: dict,
        tag: xml_tag,
        schema_tag: Tag,
        response: dict
    ):
        current_schema_item = current_schema_item[0]
        serialized_list = serialize_by_inner_schema(current_schema_item, tag)

        if not response[schema_tag.name]:
            response[schema_tag.name] = [serialized_list]
        else:
            response[schema_tag.name].append(serialized_list)



from .main import serialize_by_inner_schema
from .abstract_type import AbstractType


class String(AbstractType):
    def convert_method(self, value):
        return str(value)


class Integer(AbstractType):
    def convert_method(self, value):
        return int(value)


class Float(AbstractType):
    def convert_method(self, value):
        value = value.replace(",", ".")

        return float(value)


class Boolean(AbstractType):
    def convert_method(self, value):
        return True if value.lower() == "true" else False


class NestedType(AbstractType):
    """
    Class custom handle nested schema
    """
    schema = None
    data_handling_function = None

    def __init__(self, schema, data_handling_function=None, *args, **kwargs):
        super(NestedType, self).__init__(*args, **kwargs)
        self.schema = schema
        self.data_handling_function = data_handling_function

    def data_handle(self, data):
        if self.data_handling_function:
            return self.data_handling_function(data)

        return data

    def _serialize(self, tag):
        return serialize_by_inner_schema(self.schema, tag)

    def convert_method(self, tag):
        data = self._serialize(tag)
        response = self.data_handle(data)

        return response



from setuptools import setup


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name='xml_serializer',
    author='Illia Martyniuk',
    author_email="illymartynyk@gmail.com",
    version="1.0.2",
    url='https://github.com/Eff1c/xml_serializer',
    license='MIT',
    description='Module for serialize (convert) xml to Python dict (with Python objects)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['xml_serializer'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from abc import ABC, abstractmethod


class AbstractType(ABC):
    nullable = None

    def __init__(self, nullable=True):
        self.nullable = nullable

    @abstractmethod
    def convert_method(self, value):
        pass

    def convert(self, value):
        value = value.strip() or None

        if value is None:
            if self.nullable:
                return None

            raise TypeError("Missing required argument!")

        return self.convert_method(value)
